import csv
import io
from typing import Mapping


__all__ = ["CSVRenderer"]


class CSVRenderer:
    """Render view result to CSV.

    Views using this renderer must return a sequence of dict-like items.

    """

    def __call__(self, info):
        def _render(value: Mapping, system) -> str:
            request = system.get("request")
            if request is not None:
                response = request.response
                content_type = response.content_type
                if content_type == response.default_content_type:
                    response.content_type = "text/csv"
            rows = iter(value)
            first_row = next(rows, None)
            if first_row is None:
                return ""
            file = io.StringIO()
            fields = tuple(first_row.keys())
            header = {f: " ".join(f.split("_")).capitalize() for f in fields}
            writer = csv.DictWriter(file, fieldnames=fields)
            writer.writerow(header)
            writer.writerow(first_row)
            writer.writerows(rows)
            return file.getvalue()

        return _render



import os

here = os.path.dirname(__file__)
top_level = os.path.dirname(here)

# -- General configuration -----------------------------------------------------

project = "pyramid_resourceful"
version = "1.0a1"  # Short version
release = version  # Full version
author_name = "Wyatt Baldwin"
author_email = "self@wyattbaldwin.com"
copyright = f"2011, 2018, 2020, {author_name} <{author_email}>"
github_username = "wylee"
github_url = f"https://github.com/{github_username}/{project}"

extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
]

templates_path = ["_templates"]

source_suffix = ".rst"

master_doc = "index"

today_fmt = "%B %d, %Y"

exclude_patterns = ["_build"]

pygments_style = "sphinx"

# reStructuredText options ------------------------------------------------

# This makes `xyz` the same as ``xyz``.
default_role = "literal"

# This is appended to the bottom of all docs.
rst_epilog = """
.. |project| replace:: {project}
.. |github_url| replace:: {github_url}
""".format_map(
    locals()
)

# Options for intersphinx extension ---------------------------------------

intersphinx_mapping = {
    "python": ("https://docs.python.org/3.6", None),
}

# -- Options for HTML output -------------------------------------------

html_theme = "alabaster"

html_theme_options = {
    "description": "Resourceful routes & views for Pyramid",
    "github_user": github_username,
    "github_repo": project,
    "page_width": "1200px",
    "fixed_sidebar": True,
    "sidebar_width": "300px",
    "extra_nav_links": {
        "Source (GitHub)": github_url,
    },
}

html_sidebars = {
    "**": [
        "about.html",
        "navigation.html",
        "searchbox.html",
    ]
}

html_static_path = []

html_last_updated_fmt = "%b %d, %Y"

# Output file base name for HTML help builder.
htmlhelp_basename = f"{project}doc"


# -- Options for LaTeX output ------------------------------------------

latex_documents = [
    (
        "index",
        f"{project}.tex",
        "pyramid\\_resourceful Documentation",
        author_name,
        "manual",
    ),
]

# -- Options for manual page output ------------------------------------

man_pages = [
    (
        "index",
        project,
        f"{project} Documentation",
        [author_name],
        1,
    )
]



from . import config as config_module
from .resource import resource_config, Resource
from .view import ResourceView

__all__ = [
    "resource_config",
    "Resource",
    "ResourceView",
]


__version__ = "1.0a2"


def includeme(config):
    for name in config_module.__all__:
        method = getattr(config_module, name)
        config.add_directive(name, method)



"""SQLAlchemy ORM Example

To run this example, first run ``poetry install``, which will install
the necessary development dependencies. The run the following command
from the top level ``pyramid_resourceful`` directory::

    PYTHONPATH=. pserve -n sqlalchemy --reload examples/example.ini

Then open http://localhost:6543/ in your browser. From there, you can
play around with CRUD from a very simple UI.

A temporary SQLite database named ``example.db`` will be created in the
``examples`` directory the first time this example is run. On subsequent
runs, if this database already exists, all of its tables will be dropped
and recreated.

"""
from pyramid.config import Configurator
from pyramid.csrf import CookieCSRFStoragePolicy

from pyramid_resourceful.sqlalchemy import (
    SQLAlchemyContainerResource,
    SQLAlchemyItemResource,
)

from sqlalchemy.engine import create_engine
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import sessionmaker
from sqlalchemy.schema import Column
from sqlalchemy.types import Integer, String


Base = declarative_base()


class Item(Base):

    __tablename__ = "item"

    id = Column(Integer, primary_key=True)
    title = Column(String(100), nullable=False)
    description = Column(String)

    def __json__(self, _request):
        return {
            "id": self.id,
            "title": self.title,
            "description": self.description,
        }


class ContainerResource(SQLAlchemyContainerResource):

    model = Item


class ItemResource(SQLAlchemyItemResource):

    model = Item


def root_view(request):
    items = request.dbsession.query(Item).all()
    return {
        "items": items,
        "model": Item,
    }


def main(global_config, **settings):
    engine = create_engine(f"sqlite:///{settings['db.path']}")
    Session = sessionmaker(bind=engine)
    create_and_populate_database(engine)

    def db_session_factory(request):
        return Session()

    config = Configurator(settings=settings)

    with config:
        config.include("pyramid_mako")
        config.include("pyramid_resourceful")

        config.set_csrf_storage_policy(CookieCSRFStoragePolicy())
        config.set_default_csrf_options(require_csrf=True)

        config.add_request_method(db_session_factory, "dbsession", reify=True)

        # These are the pyramid_resourceful bits
        config.add_resource(
            ContainerResource,
            name="root",
            path="/",
            renderers=["json", "example.mako"],
        )
        config.add_resource(ContainerResource, renderers=["json", "example.mako"])
        config.add_resource(ItemResource, id_field="id")
        config.enable_post_tunneling()

    return config.make_wsgi_app()


def create_and_populate_database(engine):
    Base.metadata.drop_all(bind=engine)
    Base.metadata.create_all(bind=engine)
    engine.execute(
        Item.__table__.insert(),
        dict(title="One", description="First"),
        dict(title="Two", description="Second"),
        dict(title="Three", description="Third"),
    )



from dataclasses import dataclass, field

from pyramid.request import Request


__all__ = [
    "resource_config",
    "Resource",
]


class Resource:
    def __init__(self, request: Request):
        self.request = request

    def options(self):
        return {
            "Allow": ", ".join(self.allowed_methods),
        }


def resource_config(**view_args):
    """Specify resource config on resource method.

    Currently, all of the supplied keyword args will be passed through
    to Pyramid's ``Configurator.add_view()``. At some point, other
    resource-oriented config options might be added.

    """

    def wrapper(view_method):
        view_method.resource_config = ResourceConfig(view_args=view_args)
        return view_method

    return wrapper


@dataclass
class ResourceConfig:

    view_args: dict = field(default_factory=dict)



import json
from collections import namedtuple
from math import ceil as ceiling
from typing import Sequence

from pyramid.httpexceptions import HTTPNotFound

from sqlalchemy import inspect
from sqlalchemy.orm import joinedload
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.sql import and_, or_

from .response import exception_response
from .settings import get_setting
from .util import extract_data, get_param, NOT_SET


from .resource import Resource


class FilterSpec(namedtuple("FilterSpec", "operator value")):
    def clone(self, operator=NOT_SET, value=NOT_SET):
        if operator is NOT_SET:
            operator = self.operator
        if value is NOT_SET:
            value = self.value
        return self.__class__(operator, value)


class SQLAlchemyResource(Resource):

    joined_load_with = ()
    """Entities to load in the same query.

    Note that func:`sqlalchemy.orm.joinedload` is used to load these
    entities. For complex query logic, this might not be suitable.

    """

    filters_to_skip = ()
    """Filters that should be skipped by :meth:`apply_filters`.

    The intent behind this is to specify filters which will be handled
    specially rather than by the default filtering logic.

    """

    @property
    def key(self):
        """Key used to refer to item or items in returned data."""
        raise NotImplementedError("key property must be implemented in subclasses")

    @property
    def model(self):
        """SQLAlchemy ORM class."""
        raise NotImplementedError("model property must be implemented in subclasses")

    def __init__(self, request):
        super().__init__(request)
        model_info = inspect(self.model)
        settings = self.request.registry.settings
        self.dbsession = request.dbsession
        self.response_fields_getter = get_setting(
            settings, "get_default_response_fields", default=None
        )
        self.item_processor = get_setting(settings, "item_processor", default=None)
        self.column_attrs = tuple(attr.key for attr in model_info.column_attrs)
        self.default_response_fields = self.column_attrs

    def get_query(self):
        """Get base query.

        For more complex scenarios, this can be overridden and
        filtering, ordering, pagination, and/or joined loading can be
        disabled as needed.

        """
        return self.dbsession.query(self.model)

    def get_filters(self, *, params=None, converter=json.loads):
        """Get filters from request."""
        return {}

    def apply_filters(self, q, *, skip_filters=()):
        """Get filters and apply them to the base query.

        See :meth:`get_filters` for how filters are specified. By
        default, filters are ANDed together. This can be overridden by
        specifying `$operator=or` in the request's parameters.

        """
        filters = self.get_filters()

        if not filters:
            return q

        model = self.model
        operations = []
        boolean_operator = filters.pop("$operator", "and").lower()

        for name, spec in filters.items():
            if name in skip_filters or name in self.filters_to_skip:
                continue
            try:
                col = getattr(model, name)
            except AttributeError:
                raise exception_response(
                    400,
                    detail=f"Unknown column on model {model.__name__}: {name}",
                )
            operator = getattr(col, spec.operator)
            operations.append(operator(spec.value))

        if boolean_operator == "and":
            q = q.filter(and_(*operations))
        elif boolean_operator == "or":
            q = q.filter(or_(*operations))
        else:
            raise exception_response(
                400,
                detail=f"Unsupported boolean operator: {boolean_operator}",
            )

        return q

    def apply_options(self, q):
        """Apply options to query."""
        if self.joined_load_with:
            for item in self.joined_load_with:
                q = q.options(joinedload(item))
            q = q.populate_existing()
        return q

    def get_response_fields(self, item):
        """Get fields to include in response.

        By default, all column attributes will be included. To include
        additional fields::

            field=*&field=x&field=y&field=z

        The default fields plus ``x``, ``y``, and ``z`` will be
        included.

        To specify only some fields::

            field=a&field=b&field=c

        Only fields ``a``, ``b``, and ``c`` will be included.

        Fields can be passed via one or more ``field`` request
        parameters or* via a single ``fields`` request parameter
        formatted as a comma-separated list. These are equivalent::

            field=a&field=b&field=c
            fields=a,b,c

        """
        request = self.request
        specified = get_param(request, "field", multi=True, default=None)
        specified = specified or get_param(request, "fields", list, default=None)
        specified = specified or ["*"]
        fields = set()
        for spec in specified:
            if spec == "*":
                fields.update(self.get_default_response_fields(item))
            else:
                fields.add(spec)
        return fields

    def get_default_response_fields(self, item):
        """Get default fields to include in response."""
        model = self.model
        request = self.request
        response_fields_getter = self.response_fields_getter
        if response_fields_getter:
            return response_fields_getter(self, model, item, request)
        return self.default_response_fields

    def extract_fields(self, item, fields=None):
        """Extract fields from item.

        The incoming item is typically an ORM instance and the returned
        item is typically a dict.

        """
        request = self.request

        if fields is None:
            fields = self.get_response_fields(item)

        new_item = {}

        for name in fields:
            name, *rest = name.split(".", 1)
            obj = getattr(item, name)
            if callable(obj):
                obj = obj(request)
            if rest:
                if isinstance(obj, Sequence) and not isinstance(obj, str):
                    result = [self.extract_fields(sub_obj, rest) for sub_obj in obj]
                    if name in new_item:
                        for i, sub_obj in enumerate(result):
                            new_item[name][i].update(sub_obj)
                    else:
                        new_item[name] = result
                else:
                    result = self.extract_fields(obj, rest)
                    if name in new_item:
                        new_item[name].update(result)
                    else:
                        new_item[name] = result
            else:
                new_item[name] = obj

        return new_item

    def process_item(self, item):
        """Process item after fields have been extracted.

        The incoming item is typically a dict. By default, the item is
        returned as is.

        """
        item_processor = self.item_processor
        if item_processor:
            return item_processor(self, self.model, item, self.request)
        return item


class SQLAlchemyContainerResource(SQLAlchemyResource):

    key = "items"
    item_key = "item"

    filtering_enabled = True
    filtering_supported_operators = {
        "=": "__eq__",
        "!=": "__ne__",
        "<": "__lt__",
        "<=": "__le__",
        ">": "__gt__",
        ">=": "__ge__",
        "in": "in_",
        "not in": "notin_",
        "like": "like",
        "not like": "notlike",
        "ilike": "ilike",
        "not ilike": "notilike",
        "is": "is_",
        "is not": "isnot",
    }

    ordering_enabled = True
    ordering_default = ()

    # Enabled by default to avoid huge queries
    pagination_enabled = True
    pagination_default_page_size = 50
    pagination_max_page_size = 250

    def get(self, *, wrapped=True):
        """Get items in container."""
        data = {}
        q = self.get_query()
        if self.filtering_enabled:
            q = self.apply_filters(q)
        if self.ordering_enabled:
            q = self.apply_ordering(q)
        if self.pagination_enabled:
            q, pagination_data = self.apply_pagination(q)
            if pagination_data is not None:
                data["pagination_data"] = pagination_data
        q = self.apply_options(q)
        items = q.all()
        if not wrapped:
            return items
        items = [self.extract_fields(item) for item in items]
        items = [self.process_item(item) for item in items]
        data[self.key] = items
        return data

    def post(self):
        data = extract_data(self.request)
        item = self.model(**data)
        self.dbsession.add(item)
        self.dbsession.commit()
        return {self.item_key: item}

    def get_filters(self, *, params=None, converter=json.loads):
        """Get filters from request.

        The ``filters`` query parameter is a JSON-encoded object
        containing filter specifications using one of the following
        formats::

            1. "column": <value>
            2. "column [operator]": <value>

        In the first case, the operator will be ``=`` if the value is
        a scalar or ``in`` if the value is a list.

        For the second case, see the list of allowed operators defined
        by attr:`filtering_supported_operators`.

        Example::

            ?filters={"a": 1, "b": ["1", "2"], "c <": 4}

        This is converted to::

            a = 1 and b in ('1', '2') and c < 4

        .. note:: Filters are extracted from ``request.GET`` by default.
            Pass a different source via ``params`` if necessary (see
            :func:`get_params` for more details).

        """
        request = self.request
        filters = get_param(
            request,
            "filters",
            converter=converter,
            params=params,
            default={},
        )
        if not filters:
            return filters
        supported_operators = self.filtering_supported_operators
        processed_filters = {}
        for spec, value in filters.items():
            name, *operator = spec.split(" ", 1)
            if operator:
                operator = operator[0].lower()
            elif isinstance(value, Sequence) and not isinstance(value, str):
                operator = "in"
            else:
                operator = "="
            if operator not in supported_operators:
                raise exception_response(
                    400, detail=f"Unsupported SQL operator: {operator}"
                )
            operator = supported_operators[operator]
            processed_filters[name] = FilterSpec(operator, value)
        return processed_filters

    def apply_ordering(self, q):
        request = self.request
        ordering = get_param(request, "ordering", multi=True, default=None)
        ordering = ordering or self.ordering_default

        if not ordering:
            return q

        order_by = []
        for item in ordering:
            if isinstance(item, str):
                if item.startswith("-"):
                    item = item[1:]
                    desc = True
                else:
                    desc = False
                item = getattr(self.model, item)
                if desc:
                    item = item.desc()
            order_by.append(item)
        q = q.order_by(*order_by)

        return q

    def apply_pagination(self, q):
        request = self.request
        page = get_param(request, "page", int, default=1)

        page_size = get_param(request, "page_size", default=None)

        # XXX: Page size "*" disables pagination
        if page_size == "*":
            return q, None

        page_size = get_param(
            request, "page_size", int, default=self.pagination_default_page_size
        )

        if page < 1:
            page = 1

        if self.pagination_max_page_size and page_size > self.pagination_max_page_size:
            page_size = self.pagination_max_page_size

        count = q.count()
        num_pages = ceiling(count / page_size)
        offset = (page - 1) * page_size

        q = q.offset(offset)
        q = q.limit(page_size)

        pagination_data = {
            "pages": num_pages,
            "current_page": page,
            "previous_page": 1 if page == 1 else page - 1,
            "next_page": page + 1,
            "page_size": page_size,
            "count": count,
        }

        return q, pagination_data


class SQLAlchemyItemResource(SQLAlchemyResource):

    key = "item"

    def delete(self):
        item = self.get(wrapped=False)
        self.dbsession.delete(item)
        self.dbsession.commit()
        return {self.key: item}

    def get(self, *, wrapped=True):
        """Get an item."""
        q = self.get_query()
        q = self.apply_filters(q)
        q = self.apply_options(q)
        try:
            item = q.one()
        except NoResultFound:
            filters = self.get_filters()
            detail = f"No item found for filters: {filters!r}"
            raise exception_response(404, detail=detail)
        if not wrapped:
            return item
        item = self.extract_fields(item)
        item = self.process_item(item)
        return {self.key: item}

    def patch(self):
        item = self.get(wrapped=False)
        data = extract_data(self.request)
        for name, value in data.items():
            setattr(item, name, value)
        self.dbsession.commit()
        return {self.key: item}

    def put(self):
        try:
            item = self.get(wrapped=False)
        except HTTPNotFound:
            item = None
        data = extract_data(self.request)
        if item is None:
            item = self.model(**data)
            self.dbsession.add(item)
        else:
            # TODO: Validate that ``data`` represents a complete item?
            for name, value in data.items():
                setattr(item, name, value)
        self.dbsession.commit()
        return {self.key: item}

    def get_filters(self, *, params=None):
        request = self.request
        params = params or request.matchdict
        filters = {}
        for name, value in params.items():
            try:
                value = json.loads(value)
            except ValueError:
                pass
            filters[name] = FilterSpec("__eq__", value)
        return filters



import logging


log = logging.getLogger(__name__)


def add_cors_headers(event):
    """Add CORS headers.

    If the resource handles CORS itself by adding access control headers
    to the response, this does nothing. Otherwise...

    When the request method is OPTIONS, this will return a response that
    allows requests from the request's origin for any method and content
    type.

    When the request method is something other than OPTIONS, this will
    add a header to the response that allows the request.

    .. warning:: This is mainly intended for use in development since it
        allows CORS requests from *anywhere*, which is probably not what
        you want.

    """
    request = event.request
    response = event.response

    origin = request.headers.get("Origin", "").strip()

    if not origin:
        return

    if has_cors_headers(response):
        return

    if request.method == "OPTIONS":
        if has_cors_headers(request):
            response.headers.update(
                {
                    "Access-Control-Allow-Origin": origin,
                    "Access-Control-Allow-Methods": "*",
                    "Access-Control-Allow-Headers": "*",
                }
            )
    else:
        response.headers.update({"Access-Control-Allow-Origin": origin})


def has_cors_headers(obj):
    return any(name.startswith("Access-Control") for name in obj.headers)



from typing import Any, Callable, Dict, List

from pyramid.path import DottedNameResolver
from pyramid.settings import asbool, aslist

from .util import NOT_SET


DEFAULT_RESOURCE_METHODS = (
    "delete",
    "get",
    "options",
    "patch",
    "post",
    "put",
)


DEFAULT_SETTINGS = {
    "default_acl": None,
    "get_default_response_fields": None,
    "item_processor": None,
    "resource_methods": DEFAULT_RESOURCE_METHODS,
}
"""Default settings."""


TYPES: Dict[str, Any] = {
    "default_acl": Any,
    "get_default_response_fields": Callable[..., Any],
    "item_processor": Callable[..., Any],
    "resource_methods": List,
}
"""Types of the :data:`DEFAULT_SETTINGS`.

Used to determine how to convert settings values.

"""


def get_setting(all_settings, name, default=NOT_SET):
    """Get pyramid_resourceful setting from config.

    If the setting wasn't set in the app, the passed ``default`` value
    will be used. If a ``default`` value wasn't passed, the default from
    :data:`DEFAULT_SETTINGS` will be used.

    """
    if name not in DEFAULT_SETTINGS:
        raise KeyError(f"Unknown pyramid_resourceful setting: {name}")
    settings = all_settings.get("pyramid_resourceful", {})
    if name in settings:
        value = settings[name]
    elif default is NOT_SET:
        value = DEFAULT_SETTINGS[name]
    else:
        value = default
    if isinstance(value, str):
        type_ = TYPES[name]
        if type_ is bool:
            converter = asbool
        elif type_ is Callable[..., Any]:
            resolver = DottedNameResolver()
            converter = resolver.maybe_resolve
        elif type_ is List:
            converter = aslist
        elif type_ is Any:
            resolver = DottedNameResolver()
            converter = resolver.maybe_resolve
        else:
            converter = None
        if converter is not None:
            value = converter(value)
    return value



import logging

from pyramid.csrf import check_csrf_origin
from pyramid.httpexceptions import HTTPSeeOther, HTTPNoContent
from pyramid.request import Request


__all__ = ["ResourceView"]


log = logging.getLogger(__name__)


class ResourceView:
    def __init__(self, context, request: Request):
        self.resource = context
        self.request = request

        # XXX: This is for Pyramid, in case it needs it for something
        self.context = context

    def delete(self):
        data = self.resource.delete()
        return self.get_standard_response(data)

    def get(self):
        data = self.resource.get()
        return self.get_standard_response(data)

    def options(self):
        options = self.resource.options()
        return HTTPNoContent(headers=options)

    def patch(self):
        data = self.resource.patch()
        return self.get_standard_response(data)

    def post(self):
        data = self.resource.post()
        return self.get_standard_response(data)

    def put(self):
        data = self.resource.put()
        return self.get_standard_response(data)

    def get_standard_response(self, data):
        """Get a standard response.

        When there's no data, return a ``204 No Content`` response
        regardless of the request method or configured renderers.
        Otherwise...

        For XHR requests, return a ``200 OK`` response with rendered
        data for *all* request methods (typically JSON).

        For non-XHR requests:

        - Return a ``200 OK`` response with rendered data for GET
          requests (typically HTML rendered from a template). For HEAD
          requests, the data is also returned so that Pyramid can
          generate a HEAD response.

        - Return a ``303 See Other`` response for DELETE, PATCH, PUT,
          and POST requests:

          - For DELETEs, use the referrer if it's safe (same origin) or
            fall back to the URL of the current resource.

          - For other methods, fall back to the URL of the current
            resource (which is always safe).

        """
        if data is None:
            return HTTPNoContent()

        request = self.request
        method = request.method

        if method in ("GET", "HEAD") or request.is_xhr:
            converter = getattr(self.resource, "response_converter", None)
            if converter:
                data = converter(data)
            return data

        # Redirect after POST et al.
        if method == "DELETE":
            # XXX: It might be nice to automatically redirect to the
            # associated container resource on DELETE, but that seems
            # complicated. Redirecting back to the referrer is, so we'll
            # stick with that for that now.
            if check_csrf_origin(request, raises=False):
                location = request.referrer
            else:
                location = request.path_info
                log.warning(
                    "The referrer for this DELETE request isn't from a "
                    "trusted origin; redirecting back to the current"
                    "resource (%s) instead, which will result in a 404 "
                    "response",
                    request.matched_route.name,
                )
            return HTTPSeeOther(location=location)
        if method in ("PATCH", "POST", "PUT"):
            # Redirect back to the current resource.
            #
            # XXX: For container resources, it's debatable as to
            # whether it's preferable to redirect to the new item or
            # back to the container, but the latter doesn't require
            # any special configuration or logic.
            location = request.path_info
            return HTTPSeeOther(location=location)



from pyramid.httpexceptions import status_map


def json_formatter(self, *args, **kwargs):
    return {
        "title": self.title,
        "explanation": self.explanation,
        "detail": self.detail,
        "comment": self.comment or "",
        # Everything munged together
        "message": f"{self.explanation}\n\n{self.detail}\n\n{self.comment}".rstrip(),
    }


def exception_response(
    status_code, title=None, explanation=None, detail=None, **kwargs
):
    """Replacement for Pyramid's ``exception_response``.

    Pyramid's :func:`pyramid.httpexceptions.exception_response` munges
    the explanation, detail, and comment together into a single message
    field, but it's often nice to access these independently in JSON
    responses.

    For JSON responses, this keeps the message field but also adds
    separate explanation, detail, and comment fields.

    Pyramid also adds the HTTP status code to the JSON response object,
    which is redundant, so that's not included.

    In addition, this allows the ``title`` and ``explanation`` fields to
    be overridden.

    For HTML responses, the default Pyramid behavior is used.

    """
    error_type = status_map[status_code]
    type_dict = {
        "_json_formatter": json_formatter,
    }
    if title:
        type_dict["title"] = title
    if explanation:
        type_dict["explanation"] = explanation
    error_type = type(error_type.__name__, (error_type,), type_dict)
    return error_type(detail=detail, **kwargs)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pyramid_resourceful']

package_data = \
{'': ['*']}

install_requires = \
['pyramid']

extras_require = \
{':python_version == "3.6"': ['dataclasses']}

setup_kwargs = {
    'name': 'pyramid-resourceful',
    'version': '1.0a2',
    'description': 'Resourceful routes & views for Pyramd',
    'long_description': 'pyramid_resourceful\n+++++++++++++++++++\n\n``pyramid_resourceful`` is a somewhat-opinionated toolkit for building\nresourceful Web services and applications on top of the Pyramid Web\nframework.\n\nTake a look in the ``examples`` directory for self-contained, runnable\nexamples.\n\nSee https://pyramid-resourceful.readthedocs.io/ for detailed documentation\nof interfaces, APIs, and usage.\n\nLicense\n=======\n\nThis package is provided under the MIT license. See the ``LICENSE`` file\nfor details.\n',
    'author': 'Wyatt Baldwin',
    'author_email': 'self@wyattbaldwin.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/wylee/pyramid_resourceful',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import logging
import posixpath
from contextlib import contextmanager
from datetime import date, datetime
from decimal import Decimal
from typing import List

from pyramid.config import Configurator, ConfigurationError
from pyramid.events import NewRequest, NewResponse
from pyramid.interfaces import IRendererFactory

from .cors import add_cors_headers
from .response import exception_response
from .settings import get_setting
from .util import camel_to_underscore
from .view import ResourceView


__all__ = [
    "add_json_adapter",
    "add_json_adapters",
    "add_resource",
    "add_resources",
    "enable_cors",
    "enable_post_tunneling",
]


log = logging.getLogger(__name__)


RENDERER_EXT_MAP = {
    "jinja2": "html",
    "mako": "html",
}


RENDERER_ACCEPT_MAP = {
    "csv": "text/csv",
    "json": "application/json",
    "html": "text/html",
}


def add_json_adapter(self: Configurator, adapter):
    """Add a JSON adapter."""
    type_, adapter = adapter
    renderer = self.registry.getUtility(IRendererFactory, "json")
    renderer.add_adapter(type_, adapter)


def add_json_adapters(self: Configurator, *adapters):
    """Add default and additional JSON adapters.

    Adds default JSON adapters for date, datetime, and decimal objects:

    - date -> ISO date string
    - datetime -> ISO datetime string
    - decimal -> string

    Also adds additional adapters if specified.

    .. note:: If you don't want the defaults, use
        :func:`add_json_adapter` instead.

    """
    renderer = self.registry.getUtility(IRendererFactory, "json")
    adapters = (
        (date, lambda obj, req: obj.isoformat()),
        (datetime, lambda obj, req: obj.isoformat()),
        (Decimal, lambda obj, req: str(obj)),
    ) + adapters
    for type_, adapter in adapters:
        renderer.add_adapter(type_, adapter)


def add_resource(
    self: Configurator,
    resource_class,
    name=None,
    renderers=["json"],
    # Resource args
    acl=None,
    # Route args
    path=None,
    path_prefix=None,
    id_field=None,
    route_args=None,
    # View args
    view=ResourceView,
    permission=None,
    view_args=None,
):
    """Add routes and views for a resource.

    Given a resource (class), generates a set of routes and associated
    views.

    Args:
        resource_class: The class that implements the resource. An
            ``allowed_methods`` attribute will be attached to this
            class (if it doesn't already have that attribute).

        name: The base route name. If not specified, this will be
            computed from the resource class's module path and class
            name as follows:

            - If the class name ends with "Resource", that's stripped
              off
            - The class name is then converted to camel case
            - The module name containing the class is joined to the
              converted class name with a dot

            So, for a resource class named ``ContainerResource`` in a
            module named ``api``, the computed name will be
            "api.container".

        renderers: A list of renderers to generate routes and views for.
            This can include entries like "json" and/or "template.mako".

            For each renderer, a route with an extension (like ".json"
            and/or ".html") is generated with a view for each
            corresponding resource method. In this case, the response is
            rendered according to the extension.

            In addition, a route with *no* extension is generated with
            a view for each corresponding resource method. In this case,
            the response is rendered according to the Accept header in
            the request.

        acl: An ACL list that will be attached to the resource class as
            its ``__acl__`` attribute. If this isn't specified and the
            resource class doesn't have an ``__acl__`` attribute
            already, the default ACL will be attached instead (assuming
            a default ACL is configured).

        path: The base route path. If not specified, this will be
            computed from ``name`` by replacing dots with slashes
            and underscores with dashes. For the example above, the
            computed path would be "/api/container".

        path_prefix: If specified, this will be prepended to all
            generated route paths.

        id_field: If specified, an ID-field path segment will be
            appended to ``path``. E.g., if "id" is passed, the base path
            will be set to "/whatever/path/{id}". This is useful when an
            explicit ``path`` isn't passed.

        route_args: Common route args that will be passed to *every*
            call to ``config.add_view()``.

        view: The view class that will call methods on the resource
            class. This is optional because in many cases the included/
            default view class will suffice.

        permission: Permission that will be set on all generated views.
            This is a special case for convenience.

        view_args: Common view args that will be passed to *every* call
            to ``config.add_view()``.

    Settings can be set under the "pyramid_resourceful" key:

    - default_acl: Default ACL to attach to resource classes that
      don't have an __acl__ attribute (when ``acl`` isn't specified).

    - resource_methods: Methods on resource classes that will be
      considered resource methods (these should be lower case); if not
      specified, the default :data:`RESOURCE_METHODS` will be used.

    """
    resource_class = self.maybe_dotted(resource_class)
    resource_name = resource_class.__name__
    view = self.maybe_dotted(view)
    view_name = view.__name__

    if name is None:
        module_name = resource_class.__module__.rsplit(".", 1)[-1]
        name = resource_name
        suffix = "Resource"
        if name.endswith(suffix):
            name = name[: -len(suffix)]
        name = camel_to_underscore(name)
        name = f"{module_name}.{name}"
        log.debug(f"Computed route name: {name}")

    if path is None:
        path = name.replace(".", "/")
        path = path.replace("_", "-")
        path = path.strip("/")
        path = f"/{path}"
        log.debug(f"Computed route pattern: {path}")

    if path_prefix is not None:
        log.debug(f"Prepending path prefix {path_prefix} to route pattern: {path}")
        path = posixpath.join(path_prefix, path.lstrip("/"))

    if id_field is not None:
        log.debug(f"Appending ID field {id_field} to route pattern: {path}")
        path = posixpath.join(path, f"{{{id_field}}}")

    if acl is not None:
        resource_class.__acl__ = acl
    else:
        acl = getattr(resource_class, "__acl__", None)
        if acl is None:
            default_acl = get_setting(self.get_settings(), "default_acl")
            if default_acl is not None:
                log.debug(
                    f"Setting {resource_name}.__acl__ " f"to default ACL: {default_acl}"
                )
                resource_class.__acl__ = default_acl

    route_args = {} if route_args is None else route_args

    view_args = {} if view_args is None else view_args
    view_args.setdefault("http_cache", 0)
    if permission:
        view_args["permission"] = permission

    resource_methods = get_setting(self.get_settings(), "resource_methods")
    resource_methods = tuple(m.lower() for m in resource_methods)
    request_methods = tuple(m.upper() for m in resource_methods)

    method_config = []
    allowed_methods = []
    methods_not_allowed = []
    for method in resource_methods:
        attr = method.lower()
        request_method = method.upper()
        if hasattr(resource_class, attr):
            resource_method = getattr(resource_class, attr)
            resource_config = getattr(resource_method, "resource_config", None)
            method_config.append((method, request_method, resource_config))
            allowed_methods.append(request_method)
            if not hasattr(view, attr):
                raise ConfigurationError(
                    f"View {view_name} "
                    f"has no method `{attr}` "
                    f"corresponding to resource method `{attr}` "
                )
        else:
            methods_not_allowed.append(request_method)

    if allowed_methods:
        if not hasattr(resource_class, "allowed_methods"):
            resource_class.allowed_methods = allowed_methods
    else:
        raise ConfigurationError(
            f"No resource methods found for resource: {resource_name}"
        )

    if methods_not_allowed:
        log.debug(
            f"Resource {resource_name} does not allow these methods: "
            f"{', '.join(methods_not_allowed)}"
        )

    def add_route(route_name, pattern, accept: List[str] = None):
        log.debug(
            f"Adding route {route_name} "
            f"with pattern {pattern} "
            f"for resource {resource_name} "
            f"responding to {', '.join(request_methods)} "
            f"accepting content type {', '.join(accept) if accept else 'ANY'}"
        )
        self.add_route(
            route_name,
            pattern,
            factory=resource_class,
            request_method=request_methods,
            accept=accept,
            **route_args,
        )

    def add_views(route_name, renderer, accept: str = None):
        for attr, request_method, view_config in method_config:
            if view_config:
                args = {**view_config.view_args, **view_args}
            else:
                args = {**view_args}

            log.debug(
                f"Adding view {view_name}.{attr} "
                f"for {route_name} "
                f"responding to {request_method} "
                f"accepting content type {accept or 'ANY'} "
                f"with renderer {renderer}"
            )

            self.add_view(
                route_name=route_name,
                view=view,
                attr=attr,
                request_method=request_method,
                accept=accept,
                renderer=renderer,
                **args,
            )

        for request_method in methods_not_allowed:
            self.add_view(
                route_name=route_name,
                view=method_not_allowed_view,
                request_method=request_method,
                accept=accept,
            )

    def run():
        accepts = []

        # Add route with extension for each renderer. In this case, the
        # accepted renderer is specified by the extension in the URL
        # path (and the Accept header is ignored).
        for renderer in renderers:
            ext, accept = get_ext_and_accept_for_renderer(renderer)
            accepts.append(accept)
            route_name = f"{name}.{ext}"
            pattern = f"{path}.{ext}"
            add_route(route_name, pattern)
            add_views(route_name, renderer)

        # Add route without extension for all renderers. In this case,
        # the accepted renderer is specified by the Accept header.
        add_route(name, path, accepts)
        for accept, renderer in zip(accepts, renderers):
            add_views(name, renderer, accept)

    run()


@contextmanager
def add_resources(self: Configurator, path_prefix, add_method=None, **shared_kwargs):
    """Add multiple resources at the specified path prefix.

    Example::

        with config.add_resources("/api") as add_resource:
            add_resource(".resources.ContainerResource")
            # -> /api/resources/container

            add_resource(".resources.ItemResource")
            # -> /api/resources/item

    """
    if add_method is None:
        add_method = self.add_resource

    def add(resource_class, **kwargs):
        kwargs = {**shared_kwargs, **kwargs}
        add_method(resource_class, path_prefix=path_prefix, **kwargs)

    yield add


def enable_cors(self: Configurator):
    """Enable CORS permissively (for use in development).

    This allows CORS requests from *anywhere*, which is probably not
    what you want, other than in development.

    .. warning:: Use with CAUTION. See :func:`add_cors_headers` for
        additional info.

    """
    self.add_subscriber(add_cors_headers, NewResponse)


def enable_post_tunneling(
    self: Configurator,
    allowed_methods=("DELETE", "PATCH", "PUT"),
    param_name="$method",
    header_name="X-HTTP-Method-Override",
):
    """Allow other request methods to be tunneled via POST.

    This allows DELETE, PATCH, and PUT and requests to be tunneled via
    POST requests. The method can be specified using a parameter or a
    header.

    The name of the parameter is "$method"; it can be a query or POST
    parameter. The query parameter will be preferred if both the query
    and POST parameters are present in the request.

    The name of the header is "X-HTTP-Method-Override". If the parameter
    described above is passed, this will be ignored.

    The request method will be overwritten before it reaches application
    code, such that the application will never be aware of the original
    request method. Likewise, the parameter and header will be removed
    from the request, and the application will never see them.

    """
    allowed_methods = sorted(allowed_methods)
    disallowed_message = (
        f"Only these methods may be tunneled over POST: {allowed_methods}."
    )

    def new_request_subscriber(event):
        request = event.request
        if request.method == "POST":
            if param_name in request.GET:
                method = request.GET[param_name]
            elif param_name in request.POST:
                method = request.POST[param_name]
            elif header_name in request.headers:
                method = request.headers[header_name]
            else:
                return  # Not a tunneled request
            if method in allowed_methods:
                request.GET.pop(param_name, None)
                request.POST.pop(param_name, None)
                request.headers.pop(header_name, None)
                request.method = method
            else:
                raise exception_response(405, detail=disallowed_message)

    self.add_subscriber(new_request_subscriber, NewRequest)


def get_ext_and_accept_for_renderer(renderer):
    if "." in renderer:
        ext = renderer.rsplit(".", 1)[1]
    else:
        ext = renderer
    ext = RENDERER_EXT_MAP.get(ext, ext)
    return ext, RENDERER_ACCEPT_MAP.get(ext)


def method_not_allowed_view(request):
    raise exception_response(405)



import re
from typing import Any, List, Union

from pyramid.interfaces import ICSRFStoragePolicy, IDefaultCSRFOptions

from .response import exception_response


NOT_SET = object()
"""Represents the complete absence of a value."""


def as_bool(string: str) -> bool:
    """Convert string to bool.

    Only the values "1", "true", "0", and "false" are accepted.

    """
    string = string.lower()
    if string in ("1", "true"):
        return True
    if string in ("0", "false"):
        return False
    raise ValueError('Expected value to be one of "1", "true", "0", or "false"')


def as_list(string, sep=",", strip=True):
    """Convert string to list, splitting on comma by default."""
    string = string.strip()
    items = string.split(sep)
    if strip:
        items = [item.strip() for item in items]
    return items


def camel_to_underscore(name):
    """Convert camel case name to underscore name."""
    name = re.sub(r"(?<!\b)(?<!_)([A-Z][a-z])", r"_\1", name)
    name = re.sub(r"(?<!\b)(?<!_)([a-z])([A-Z])", r"\1_\2", name)
    name = name.lower()
    return name


def extract_data(request):
    """Extract request data."""
    content_type = request.content_type

    if content_type == "application/x-www-form-urlencoded":
        data = request.POST
    elif content_type == "application/json":
        data = request.json_body if request.body else None
    else:
        raise TypeError(f"Cannot extract data for content type: {content_type}")

    # Remove CSRF token from data, if present
    policy = request.registry.queryUtility(ICSRFStoragePolicy)
    if policy:
        options = request.registry.queryUtility(IDefaultCSRFOptions)
        if options:
            token = options.token
            if token in data:
                del data[token]

    return data


def get_param(
    request,
    name,
    converter=None,
    params=None,
    *,
    multi=False,
    strip=False,
    convert_blank_to_none=True,
    default=NOT_SET,
) -> Union[Any, List[Any]]:
    """Get the specified request parameter and, optionally, convert it.

    Args:
        params: Any dict-like object that *also* has ``getone`` and
            ``getall`` methods. Defaults to ``request.GET``.

        multi (bool): If set, all params with ``name`` will be collected
            into a list (even if there's only one such param). If not
            set, only one param with ``name`` may be present in
            ``params``.

        strip (bool): If set, param values will be stripped before being
            converted.

        convert_blank_to_none (bool): If set, blank param values will be
            converted to ``None``. Blank param values look like ``a=``.

        converter (callable): If specified, the value or values will be
            passed to this callable for conversion (unless converted the
            value was already converted to ``None``). If a value can't
            be parsed, a 400 exception response will be raised.

        default (any): A default value to return when the param isn't
            present. If the param isn't present and no default value is
            specified, a ``KeyError`` will be raised.

    Returns:
        any: When ``multi`` is not set
        list[any]: When ``multi`` is set

    Raises:
        KeyError: Param isn't present and no default value is given.
        KeyError: ``multi=False`` but param is present multiple times.
        400 response: Param value can't be converted by ``converter``.

    Special handling of flag parameters:

    If ``converter`` is ``bool`` or :func:`as_bool` *and* the param is
    specified only once *and* the param has *no* value, the param will
    be handled as a flag. Note that in this example, ``c`` is *not*
    considered a flag due to the equal sign (it has a *blank* value)::

        >>> from pyramid.request import Request
        >>> req = Request.blank("/endpoint?a&!b&c=")
        >>> get_param(req, "a", bool) -> True
        >>> get_param(req, "b", bool) -> False

    """
    if params is None:
        params = request.GET

    if converter is bool:
        converter = as_bool
    elif converter is list:
        converter = as_list

    # Special handling of flags.
    if converter is as_bool and not multi:
        maybe_flags = []
        # NOTE: Splitting is based on urllib.parse.parse_qsl()
        for group in request.query_string.split("&"):
            group_params = group.split(";")
            for param in group_params:
                if "=" not in param:
                    if param.startswith("!"):
                        if param[1:].strip():
                            maybe_flags.append(param)
                    elif param.strip():
                        maybe_flags.append(param)
        if maybe_flags:
            count = maybe_flags.count(name)
            negated_count = maybe_flags.count(f"!{name}")
            if count == 1 and negated_count == 0:
                return True
            if count == 0 and negated_count == 1:
                return False

    if name not in params:
        if default is NOT_SET:
            raise KeyError(f"Param not present: {name!r}")
        return default

    def convert(v):
        if strip:
            v = v.strip()
        if not v and convert_blank_to_none:
            return None
        if converter:
            try:
                v = converter(v)
            except (TypeError, ValueError):
                raise exception_response(
                    400,
                    detail=f"Could not parse parameter {name} with {converter}: {v!r}",
                )
        return v

    if multi:
        values = params.getall(name)
        values = [convert(value) for value in values]
        return values

    value = params.getone(name)
    value = convert(value)
    return value
